/*
	Ray's abilities by SMS Alfredo
	Featuring Snu
	
	Code snippets borrowed from sloperollangle by Golden
	https://mb.srb2.org/threads/sloperollangle-character-rotation-on-slopes.27164/
	
	This Lua also references Ray's ability code from the Sonic Mania
	Decompilation in order to recreate certain gliding functionality.
	https://github.com/Rubberduckycooly/Sonic-Mania-Decompilation
	
	Usages of both are labelled via comments and free to use by anyone,
	as I did not make nor own the original code.
	If you're unsure what is what, please ask!
*/

rawset(_G, "Ray", {})

//
//Sounds
//

for i, sound in ipairs({"sfx_raydv1", "sfx_raydv2", "sfx_raypl1", "sfx_raypl2"})
	freeslot(sound)
	local sfx = _G[sound]
	sfxinfo[sfx].caption = "/"
end

freeslot("sfx_raygst")
sfxinfo[sfx_raygst].caption = "Whoosh"

//
//Constants
//

Ray.SKIN = "ray"

//Sound constants, when the loop should start and end
Ray.GLIDESTARTSOUND = 46
Ray.GLIDELOOPSOUND = 118

//Angles the glide can be at
Ray.GLIDEMINANGLE = ANGLE_22h
Ray.GLIDEMAXANGLE = (ANGLE_90 - Ray.GLIDEMINANGLE) + ANGLE_90

//Values relating to turning
Ray.GLIDEINPUTTHRESHOLD = 35 //How far you need to hold forward
Ray.GLIDETURNFACTOR = 16 //How fast we turn horizontally
Ray.GLIDEPITCHFACTOR = ANGLE_11hh*3/2 //How fast we turn vertically

//Speed gain/loss during glide
Ray.GLIDEVERTICALSPEED = FRACUNIT
Ray.GLIDEHORIZONTALSPEED = FRACUNIT*2/3
Ray.GLIDEPULLUPFACTOR = FRACUNIT*3/2 //Multiplier to losing speed

//Speed cut when landing after glide
Ray.GLIDELANDDIV = {4,5}

//How fast the glide degrades
Ray.GLIDEMAXPOWER = 256
Ray.GLIDEDEGRADE = Ray.GLIDEMAXPOWER / 8

//Different values for the rise calculations
Ray.GLIDERISESHIFT = 10 //Horizontal speed shift
Ray.GLIDERISEMOMZ = 64*FRACUNIT //Vertical speed maximum for sine curve
Ray.GLIDERISEMOMZCAP = 12*FRACUNIT //Additive vertical speed cap
Ray.GLIDERISEDIV = 3 //Vertical speed division factor
Ray.GLIDERISECAP = 24*FRACUNIT //Cap for the rise as a whole (only really matters for spindash glide in ACZ)
Ray.GLIDERISETIMER = TICRATE //Max time the rise can last for (in cases when the rise gets stuck)

-- the snu touches
Ray.GLIDEEASEFUNC = ease.incubic	-- the easing function to use to curve the scalar
Ray.GLIDEMAXSPEED = FRACUNIT*80	-- when the scalar will max out (it'll keep increasing from here)
Ray.GLIDEMAXSCALAR = FRACUNIT*74/100	-- what value the scalar will be at when it maxes out
Ray.GLIDEUNDERWATERMULTIPLIER = FRACUNIT*2/3	-- for glide minspeed
Ray.GLIDEBATTLEMULTIPLIER = FRACUNIT*5/9	-- for the max speed (in battle)

//Values relating to glide visuals
Ray.GLIDESPR2 = SPR2_GLID
Ray.GLIDEGUSTSPR2 = SPR2_FIRE
Ray.GLIDEGUSTTIME = TICRATE //How long the gust lasts visually at a minimum
Ray.GLIDEGUSTDELAY = 4 //Frame speed and vertical boost delay of gust

//Minimum speed
Ray.GLIDEMINDIVE = FRACUNIT*-2

//
//Speed Trail (taken from Chaotix)
//

if not SPR_CXTS
	freeslot("SPR_CXTS", "S_CHAOTIX_SPEEDTRAIL")
	local fuse = TICRATE/4
	A_ChaotixSpeedTrail = function(mo)
		if not mo.fuse then mo.fuse = fuse end
		local scale = ease.outsine(FRACUNIT*mo.fuse/fuse, FRACUNIT)
		mo.spritexscale = scale + FixedMul(FRACUNIT - scale, abs(sin(mo.rollangle)))
		mo.spriteyscale = scale + FixedMul(FRACUNIT - scale, abs(cos(mo.rollangle)))
	end
	states[S_CHAOTIX_SPEEDTRAIL] = {SPR_CXTS, A|FF_PAPERSPRITE, 1, A_ChaotixSpeedTrail, 0, 0, S_CHAOTIX_SPEEDTRAIL}
end

Ray.SpawnSpeedTrail = function(mo, condition, v)
	if condition == 0 or condition == false then return end -- set condition
	if v == nil then v = {} end
	
	-- get momentum
	if v.momx == nil then v.momx = mo.momx end
	if v.momy == nil then v.momy = mo.momy end
	if v.momz == nil then v.momz = mo.momz end
	
	-- set up the random positioning
	local hdist = (v.radius or mo.radius*2)/mo.scale
	local vdist = (v.height or mo.height)/mo.scale
	vdist = P_RandomRange(0, $)*FRACUNIT
	
	-- get the angle
	local angle = mo.angle
	if v.momx or v.momy
		angle = R_PointToAngle2(0,0, v.momx, v.momy)
	elseif mo.player
		angle = mo.player.drawangle
	end
	
	-- spawn the trail
	local thok = P_SpawnMobjFromMobj(mo,
	P_RandomRange(-hdist, hdist)*FRACUNIT,
	P_RandomRange(-hdist, hdist)*FRACUNIT,
	vdist, MT_THOK)
	
	-- set the angles and state
	thok.angle = angle
	thok.rollangle = R_PointToAngle2(0,0, FixedHypot(v.momx, v.momy), v.momz*P_MobjFlip(mo))
	thok.state = S_CHAOTIX_SPEEDTRAIL
	
	-- set scale and color
	if v.scale
		local oldheight = thok.height
		thok.scale = v.scale
		if mo.eflags & MFE_VERTICALFLIP
			P_SetOrigin(thok, thok.x, thok.y, thok.z + (oldheight - thok.height))
		end
	end
	if v.color
		thok.color = v.color
		thok.colorized = true
	end
	
	-- thrust back a little
	local factor = -4*thok.scale
	P_Thrust(thok, angle, FixedMul(abs(cos(thok.rollangle)), factor))
	P_SetObjectMomZ(thok, FixedMul(sin(thok.rollangle), factor))
end

//
//Functions
//

//Is valid and Ray
Ray.IsRay = function(player)
	local mo = player and player.mo
	return mo and mo.valid and mo.skin == Ray.SKIN
end

//Can't move conditions
Ray.CantMove = function(player)
	local mo = player and player.mo
	return player.charability != CA_GLIDEANDCLIMB
	or P_PlayerInPain(player)
	or player.pflags & PF_SLIDING
	or player.powers[pw_carry]
	or player.spectator
	or player.playerstate
	or not mo.health
	or player.powers[pw_super] == 1
	or player.iseggrobo
	or player.isjettysyn
	or player.guard
	or player.actionstate
	or player.powers[pw_nocontrol]
	or (player.isxmomentum and mo.state == S_PLAY_FACEPLANT)
end

//Start/Stop glide sounds
//If sound exists, play that sound
//If sound is false, cut out only the dive sounds
//If sound is true, cut out only the pull sounds
Ray.GlideSound = function(mo, sound)
	if sound != true
		S_StopSoundByID(mo, sfx_raydv1)
		S_StopSoundByID(mo, sfx_raydv2)
	end
	
	if sound != false
		S_StopSoundByID(mo, sfx_raypl1)
		S_StopSoundByID(mo, sfx_raypl2)
	end
	
	if sound and sound != true then S_StartSound(mo, sound) end
end

//Clean up after the glide
Ray.ResetGlide = function(player)
	local mo = player.mo
	if player.rayglidethrust
		if mo and mo.valid
			Ray.GlideSound(mo)
			mo.rollangle = 0
			if mo.state == S_PLAY_GLIDE
				mo.state = S_PLAY_SPRING
			end
		end
		if skins[player.skin].name == Ray.SKIN
			player.thrustfactor = skins[player.skin].thrustfactor
			player.charflags = $ | (skins[player.skin].flags & SF_SUPER)
		end
		player.rayglidethrust = false
	end
end

//Dust Ring
Ray.DustRing = function(mo)
	//Ring effect
	local ring = P_SpawnMobjFromMobj(mo, 0,0,0, MT_THOK)
	ring.fuse = mo.tics
	ring.sprite = SPR_STAB
	ring.flags2 = $ | MF2_SPLAT
	
	ring.colorized = true
	ring.color = SKINCOLOR_SUPERSILVER1
	
	ring.scalespeed = FixedMul($, ring.scale)
	ring.scale = $/8
	ring.destscale = $*16
	
	ring.height = ring.scale
	ring.target = mo
	A_CapeChase(ring)
	
	//Dust effect
	local dust = P_SpawnMobjFromMobj(mo, 0,0,0, MT_DUST)
	dust.scale = $*3
	A_CanarivoreGas(dust, MT_DUST)
	dust.scale = mo.scale
end

-- trigger landing lag knuckles style
Ray.TriggerLandingLag = function(p)
	local mo = p.mo
	if (mo.state == S_PLAY_GLIDE_LANDING) then return end
	
	P_ResetPlayer(p)
	mo.state = S_PLAY_GLIDE_LANDING
	p.pflags = $|(PF_STASIS)
	if (p.speed > FixedMul(p.runspeed, mo.scale))
		p.skidtime = $ + mo.tics
	end
	mo.momx = ((mo.momx - p.cmomx)/2) + p.cmomx
	mo.momy = ((mo.momy - p.cmomy)/2) + p.cmomy
	if (p.powers[pw_super])
		P_Earthquake(mo, mo, 256*FRACUNIT)
		S_StartSound(mo, sfx_s3k49)
	else
		S_StartSound(mo, sfx_s3k4c)
	end
end

//
//Hooks
//

//Start the glide
Ray.AbilitySpecial = function(player)
	local mo = player.mo
	if not (Ray.IsRay(player) and player.charability == CA_GLIDEANDCLIMB)
	then return end
	if Ray.CantMove(player) or player.climbing
	or player.pflags & PF_JUMPSTASIS or (player.pflags & PF_THOKKED
	and not (player.charflags & SF_MULTIABILITY))
	then return true end
	
	//Reverse the input angle if we're going away from the camera
	player.rayglideangle = mo.angle
	if player.pflags & PF_ANALOGMODE
	and not (mo.flags2 & MF2_TWOD or twodlevel)
	and (start or player.cmd.forwardmove or player.cmd.sidemove)
		player.rayglideangle = player.cmd.angleturn<<16
		if player.cmd.forwardmove or player.cmd.sidemove
			player.rayglideangle = $ + R_PointToAngle2(0,0,
			abs(player.cmd.forwardmove)*FRACUNIT, -player.cmd.sidemove*FRACUNIT)
		end
	end
	
	//Start going up or down based on our forward input
	local forwardmove = player.cmd.forwardmove
	if mo.flags2 & MF2_TWOD or twodlevel
		forwardmove = player.cmd.sidemove
		if player.rayglideangle < 0
			forwardmove = -$
		end
	end
	
	//Get speed, is relative to your speed and the angle change
	local minspeed = FixedMul(player.actionspd, mo.scale)
	if (mo.eflags & MFE_UNDERWATER) then minspeed = FixedMul(Ray.GLIDEUNDERWATERMULTIPLIER, $) end
	local oldspeed = FixedHypot(mo.momx, mo.momy)
	P_Thrust(mo, player.rayglideangle, minspeed)
	local newspeed = FixedHypot(mo.momx, mo.momy)
	
	//Set speed
	local speed = max(min(oldspeed, newspeed), minspeed)
	P_InstaThrust(mo, player.rayglideangle, speed)
	
	//Set glide power and rise (Mania Decomp: Player_JumpAbility_Ray)
	if not player.glidetime then player.rayglidepower = Ray.GLIDEMAXPOWER end

	if forwardmove <= 0
		//Gain height from speed (Mania Decomp: Player_State_RayGlide)
		player.raygliderise = (player.rayglidepower * ((speed >> 1) + (speed >> 2) + (speed >> 4))) >> Ray.GLIDERISESHIFT
		
		//Cap rise
		player.raygliderise = min($, Ray.GLIDERISECAP / (mo.eflags & MFE_UNDERWATER and 3 or 1))
		player.fly1 = Ray.GLIDERISETIMER
		
		//Decrease power (Mania Decomp: Player_State_RayGlide)
		if player.rayglidepower > Ray.GLIDEDEGRADE
			player.rayglidepower = max(0, $ - Ray.GLIDEDEGRADE)
		end
		
		//Turn camera to angle with no input
		if not (player.cmd.forwardmove or player.cmd.sidemove)
			mo.angle = $
		end 
		player.rayglideup = true
		player.rayglideframe = skins[player.skin].sprites[mo.sprite2].numframes - 1
	else
		player.rayglideup = false
		player.rayglideframe = 0
	end
	
	mo.momz = $/2 //Mania Decomp: Player_JumpAbility_Ray
	player.rayglidepitch = ANGLE_90 //Mania Decomp: Player_JumpAbility_Ray
	player.rayglidepitch2 = ANGLE_90
	player.rayglideprevangle = player.rayglideangle
	
	mo.state = S_PLAY_GLIDE
	player.pflags = $ | PF_THOKKED | PF_NOJUMPDAMAGE | PF_CANCARRY
	player.pflags = $ &~ PF_APPLYAUTOBRAKE &~ PF_STARTJUMP &~ PF_SPINNING
	player.rayrollangle = mo.rollangle
	player.rayglidesound = 0
	
	player.drilltimer = 0
	player.secondjump = 0
	player.glidetime = max($, 1)
	return true
end

//Glide puff, like the one from Extra Slot Ray
Ray.JumpSpinSpecial = function(player)
	local mo = player.mo
	if not Ray.IsRay(player) or Ray.CantMove(player) or player.climbing
	or player.pflags & (PF_JUMPSTASIS|PF_SPINDOWN) or player.secondjump
	or player.glidetime != 1 or mo.state != S_PLAY_GLIDE
	then return end
	
	//Set rise to jump momentum
	local oldmomz = mo.momz
	player.pflags = $ &~ PF_JUMPED
	P_DoJump(player, false)
	player.raygliderise = mo.momz*P_MobjFlip(mo)*4/5
	player.fly1 = Ray.GLIDERISETIMER
	mo.momz = oldmomz/2
	if mo.momz*P_MobjFlip(mo) > 0
		player.raygliderise = max($ + mo.momz*P_MobjFlip(mo), oldmomz*P_MobjFlip(mo))
	end
	if (mo.eflags & MFE_UNDERWATER) then player.raygliderise = $*4/5 end
	
	//Glide up!
	player.rayglideup = true
	
	player.drilltimer = Ray.GLIDEGUSTTIME
	player.secondjump = 2
	player.glidetime = $ + 1
	return true
end

//Get ability back from damaging other things
Ray.Damage = function(mobj, mo)
	local player = mo and mo.valid and mo.player
	if Ray.IsRay(player) and not Ray.CantMove(player)
	and mo.state != S_PLAY_GLIDE and player.glidetime
	and player.pflags & PF_JUMPED and player.pflags & PF_THOKKED
		player.pflags = $ &~ PF_THOKKED
		player.glidetime = 0
	end
end

//During abilities
Ray.AbilityThinker = function(player)
	local mo = player.mo
	if not Ray.IsRay(player) then return end //Gotta be Ray!
	
	//Faster stand animation
	if mo.state == S_PLAY_STND
		if not (mo.frame & FF_ANIMATE) then mo.frame = $ | FF_ANIMATE end
		mo.anim_duration = min($, 4)
	end
		
	//Wait animation
	if mo.state == S_PLAY_WAIT
		//Don't speed up the first frame
		if not mo.raystartedwait
			if (mo.frame & FF_FRAMEMASK)
				mo.raystartedwait = true
				mo.anim_duration = min($, 4)
			end
			
		//Speed up transitional frames
		elseif ((mo.frame & FF_FRAMEMASK) % G) != D
			mo.anim_duration = min($, 4)
		end
	elseif mo.raystartedwait
		mo.raystartedwait = false
	end
	
	//Can't move!
	if Ray.CantMove(player) then Ray.ResetGlide(player) return end
	
	//During glide
	if player.glidetime and mo.state == S_PLAY_GLIDE
	and not player.climbing
		//Cancel glide
		if not (player.cmd.buttons & BT_JUMP)
		or (player.gotflag or player.gotcrystal)
			player.pflags = $ &~ PF_NOJUMPDAMAGE
			mo.state = S_PLAY_JUMP
			mo.raylandinglag = (CBW_Battle) and 2 or 0
			return
		end
		player.raygroundcut = true
		
		//Signify zero'd out thrustfactor and match momentum angle after
		if not player.rayglidethrust
			player.rayglidethrust = true
		elseif player.rmomx or player.rmomy
			player.rayglideangle = R_PointToAngle2(0,0, player.rmomx, player.rmomy)
		end
		player.thrustfactor = 0
		
		//Prevent turning super
		if not player.powers[pw_super]
			player.charflags = $ &~ SF_SUPER
		end
		
		//Change what's forward in 2D
		local forwardmove = player.cmd.forwardmove
		if mo.flags2 & MF2_TWOD or twodlevel
			forwardmove = player.cmd.sidemove
			if player.rayglideangle < 0
				forwardmove = -$
			end
		end
		
		//Proper gravity
		local glidegrav = P_GetMobjGravity(mo)
		if P_IsObjectInGoop(mo)
			glidegrav = $/-2
		end
		
		//Speed and flags
		local speed = FixedHypot(mo.momx, mo.momy)
		local actionspd = player.actionspd
		player.powers[pw_noautobrake] = max($, 2)
		
		//Turning
		if not (mo.flags2 & MF2_TWOD) and not twodlevel
			local sidemove = player.cmd.sidemove
			local auto = not (player.pflags & PF_ANALOGMODE) and not sidemove
			local olddir = player.rayglideangle
			local diff = player.rayglideangle - mo.angle
			if auto and player.rayglideangle != mo.angle
				if diff > 0
					sidemove = 50
				else
					sidemove = -50
				end
			end
			
			local change = Ray.GLIDETURNFACTOR*sidemove*FRACUNIT
			player.rayglideangle = $ - change
			
			if auto
				local newdiff = player.rayglideangle - mo.angle
				if (diff < 0 and newdiff > 0)
				or (diff > 0 and newdiff < 0)
					player.rayglideangle = mo.angle
				end
			end
			P_InstaThrust(mo, player.rayglideangle, speed)
		end
		player.drawangle = player.rayglideangle
		
		//Visual Turning with rollangle
		local diff = player.rayglideprevangle - player.rayglideangle
		local limit = max(ANGLE_22h, abs(diff)/2)
		if abs(diff) > ANG1
			local change = ANG2*2
			if diff > 0 then change = -$ end
			player.rayrollangle = min(max($ + change, -limit), limit)
		else
			player.rayrollangle = FixedMul($, FRACUNIT*7/8)
		end
		
		//Calculate angle to rollangle (thanks Golden for sloperollangle)
		if not splitscreen
			local cameraangle = R_PointToAngle(mo.x, mo.y)
			local diff = cameraangle - player.rayglideangle
			mo.rollangle = FixedMul(player.rayrollangle, cos(diff))
		end
		player.rayglideprevangle = player.rayglideangle
		
		//Vertical pitch (Mania Decomp: Player_State_RayGlide)
		if player.rayglideup
			player.rayglidepitch = min($ + Ray.GLIDEPITCHFACTOR, Ray.GLIDEMAXANGLE)
			if player.rayglidepitch < 0 then player.rayglidepitch = Ray.GLIDEMAXANGLE end
			
			player.rayglidepitch2 = min($ + Ray.GLIDEPITCHFACTOR/3, Ray.GLIDEMAXANGLE)
			if player.rayglidepitch2 < 0 then player.rayglidepitch2 = Ray.GLIDEMAXANGLE end
		else
			player.rayglidepitch = max($ - Ray.GLIDEPITCHFACTOR, Ray.GLIDEMINANGLE)
			player.rayglidepitch2 = max($ - Ray.GLIDEPITCHFACTOR/3, Ray.GLIDEMINANGLE)
		end
		
		//Vertical momentum (Mania Decomp: Player_State_RayGlide)
		if player.raygliderise and player.rayglideup
			//Add rise momentum
			if player.secondjump == 2
				if player.fly1 <= Ray.GLIDERISETIMER - Ray.GLIDEGUSTDELAY
					mo.momz = $ + (player.raygliderise >> 1)*P_MobjFlip(mo)
				end
			else
				mo.momz = $ + (player.raygliderise >> 2)*P_MobjFlip(mo)
			end
			
			//End rise
			player.fly1 = max(0, $ - 1)
			if mo.momz*P_MobjFlip(mo) > player.raygliderise
				mo.momz = player.raygliderise*P_MobjFlip(mo)
				player.raygliderise = 0
			elseif not player.fly1
			or mo.z <= mo.floorz or mo.z + mo.height >= mo.ceilingz
				player.raygliderise = 0
			end
		else
			local vertical = FixedMul(sin(ANGLE_112h - player.rayglidepitch), Ray.GLIDEVERTICALSPEED)
			mo.momz = $ + FixedMul(vertical, glidegrav)
		end
		
		//During up puff
		if player.secondjump == 2
			if not player.raygliderise
				player.secondjump = 0
			elseif player.fly1 == Ray.GLIDERISETIMER - Ray.GLIDEGUSTDELAY
				Ray.DustRing(mo)
				S_StartSound(mo, sfx_raygst) //sfx_s3ka0
			end
		
		//Change angle (Mania Decomp: Player_State_RayGlide)
		elseif player.rayglideup
		and player.rayglidepitch == Ray.GLIDEMAXANGLE and forwardmove > Ray.GLIDEINPUTTHRESHOLD
			player.rayglideup = false
			player.drilltimer = 0
			
		//Pull up and gain height (Mania Decomp: Player_State_RayGlide)
		elseif not player.rayglideup and forwardmove < -Ray.GLIDEINPUTTHRESHOLD
			player.rayglideup = true
			
			//Gain height (Mania Decomp: Player_State_RayGlide)
			if mo.momz*P_MobjFlip(mo) < FixedMul(Ray.GLIDEMINDIVE, mo.scale)
			or player.rayglidepower >= Ray.GLIDEMAXPOWER
				//Gain height from speed (Mania Decomp: Player_State_RayGlide)
				player.raygliderise = (player.rayglidepower * ((speed >> 1) + (speed >> 2) + (speed >> 4))) >> Ray.GLIDERISESHIFT
				
				//Gain height from vertical momentum
				local momz = FixedDiv(abs(mo.momz), mo.scale)
				momz = $ - ((FixedDiv(speed, mo.scale) - actionspd) >> Ray.GLIDERISESHIFT)
				momz = FixedDiv(min(mo.momz*P_MobjFlip(mo), Ray.GLIDERISEMOMZ/Ray.GLIDERISEMOMZCAP), Ray.GLIDERISEMOMZ)
				momz = ease.insine(momz, Ray.GLIDERISEMOMZ)
				momz = min(max($/Ray.GLIDERISEDIV, 0), Ray.GLIDERISEMOMZCAP)
				player.raygliderise = $ + FixedMul(momz, mo.scale)
				
				//Cap rise
				player.raygliderise = min($, Ray.GLIDERISECAP / (mo.eflags & MFE_UNDERWATER and 3 or 1))
				player.fly1 = Ray.GLIDERISETIMER
				
				//Decrease power (Mania Decomp: Player_State_RayGlide)
				player.rayglidepower = max(0, $ - Ray.GLIDEDEGRADE)
			end
		end
		
		//Horizontal Momentum
		local angle = R_PointToAngle2(0,0, mo.momx, mo.momy)
		if not speed then angle = player.rayglideangle end
		
		local thrust = FixedMul(sin(ANGLE_112h - player.rayglidepitch), Ray.GLIDEHORIZONTALSPEED) //Mania Decomp: Player_State_RayGlide
		if thrust < 0 
			thrust = FixedMul($, Ray.GLIDEPULLUPFACTOR)
		elseif (not player.powers[pw_super])	-- no nerf while super!
			local maxspeed = Ray.GLIDEMAXSPEED
			if (CBW_Battle) then maxspeed = FixedMul(Ray.GLIDEBATTLEMULTIPLIER, $) end
			local maxscalar = Ray.GLIDEMAXSCALAR
			local func = Ray.GLIDEEASEFUNC
			local scalar = FRACUNIT + Ray.GLIDEEASEFUNC(FixedDiv(FixedDiv(speed, mo.scale), maxspeed), maxscalar)
			
			-- scale down the thrust
			thrust = FixedDiv($, scalar)
		end
		if mo.eflags & MFE_UNDERWATER then thrust = $/2 end
		thrust = FixedMul($, mo.scale)
		
		local thrust_minspeed = actionspd
		if (mo.eflags & MFE_UNDERWATER) then thrust_minspeed = FixedMul(Ray.GLIDEUNDERWATERMULTIPLIER, $) end
		
		thrust = max($ + speed, FixedMul(thrust_minspeed, mo.scale))
		if mo.eflags & MFE_UNDERWATER
		and thrust > FixedMul(player.normalspeed*5/6, mo.scale)
			thrust = $*63/64
		end
		P_InstaThrust(mo, angle, thrust)
		
		//Speed trails when going fast enough
		if not player.rayglideup or mo.momz*P_MobjFlip(mo) > 0
			thrust = FixedHypot($, mo.momz*2)
		end
		mo.rayglidestorethrust = thrust	-- gonna use this in battlemod!!
		local runspeed = FixedMul(player.runspeed, mo.scale)*3/2
		Ray.SpawnSpeedTrail(mo, thrust > runspeed and (leveltime & 1))
		
		//Get the glide frame we should be based on the vertical pitch / gust
		local numframes = skins[player.skin].sprites[Ray.GLIDESPR2].numframes - 1
		local frame = A
		
		//Gust timer
		if player.drilltimer
			if not (player.drilltimer == 1 and mo.momz*P_MobjFlip(mo) >= 0)
				player.drilltimer = $ - 1
			end
			frame = (skins[player.skin].sprites[Ray.GLIDEGUSTSPR2].numframes - 1)/2
			local increase = (Ray.GLIDEGUSTTIME - player.drilltimer)/Ray.GLIDEGUSTDELAY
			frame = -min(increase, frame)
			
		//Vertical pitch
		else
			frame = FixedDiv(player.rayglidepitch2 - Ray.GLIDEMINANGLE, Ray.GLIDEMAXANGLE)
			frame = ease.linear($, numframes - 1) + 1
			frame = numframes - frame
		end
		
		//Move to that frame from the frame we're on
		if player.rayglideframe != frame
			if player.rayglideframe < frame
				if not player.rayglideframe
					player.rayglideframe = min(frame, C) //Shortcut out of gust
				else
					player.rayglideframe = $ + 1 //Increase frame (go down)
				end
			elseif player.rayglideframe == C and player.drilltimer
				player.rayglideframe = A //Shortcut into gust
			else
				player.rayglideframe = $ - 1 //Decrease frame (go up)
			end
		end
		frame = player.rayglideframe
		
		//Get wind/gust frames if needed and set the frame!
		local factor = (runspeed*2 - thrust)/(mo.scale*4)
		factor = min(max($, 1), 4)
		local wind = (leveltime / factor) & 1
		if frame >= B
			mo.sprite2 = Ray.GLIDESPR2
			if not wind
				if frame == B
					frame = A
				elseif frame == numframes - 1
					frame = numframes
				end
			end
		else
			mo.sprite2 = Ray.GLIDEGUSTSPR2
			frame = $*-2
			if wind then frame = $ | 1
			else frame = $ &~ 1 end
		end
		mo.frame = ($ &~ FF_FRAMEMASK) | frame
		mo.tics = -1
		
		//Start a sound from the start
		if player.rayglideup and player.rayglidesound <= 0
			player.rayglidesound = 1
			Ray.GlideSound(mo, player.rayglideup)
			S_StartSound(mo, sfx_raypl1)
			
		elseif not player.rayglideup and player.rayglidesound >= 0
			player.rayglidesound = -1
			Ray.GlideSound(mo, player.rayglideup)
			S_StartSound(mo, sfx_raydv1)
			
		//Pull back sounds
		elseif player.rayglidesound > 0
			player.rayglidesound = $ + 1
			
			//Restart sound at loop
			if player.rayglidesound >= Ray.GLIDELOOPSOUND
				Ray.GlideSound(mo, sfx_raypl2)
				player.rayglidesound = Ray.GLIDESTARTSOUND
				
			//Don't immediately stop the dive sounds so as to transition them
			elseif player.rayglidesound == Ray.GLIDESTARTSOUND/4
				Ray.GlideSound(mo, not player.rayglideup)
			end
			
		//Diving sounds
		else
			player.rayglidesound = $ - 1
			
			//Restart sound at loop
			if -player.rayglidesound >= Ray.GLIDELOOPSOUND
				Ray.GlideSound(mo, sfx_raydv2)
				player.rayglidesound = -Ray.GLIDESTARTSOUND
				
			//Don't immediately stop the pull sounds so as to transition them
			elseif -player.rayglidesound == Ray.GLIDESTARTSOUND/4
				Ray.GlideSound(mo, not player.rayglideup)
			end
		end
		
	//Reset glide
	else
		Ray.ResetGlide(player)
		
		//Lose speed when hitting the ground if we're going too fast
		if player.raygroundcut
			if (P_IsObjectOnGround(mo) or mo.eflags & MFE_JUSTHITFLOOR)
				local speed = FixedHypot(mo.momx, mo.momy)
				local normalspeed = FixedMul(player.normalspeed, mo.scale)
				if speed > normalspeed
					speed = max($*Ray.GLIDELANDDIV[1]/Ray.GLIDELANDDIV[2], normalspeed)
					P_InstaThrust(mo, R_PointToAngle2(0,0, mo.momx, mo.momy), speed)
				end
				player.raygroundcut = false
				
				//Allow buffering a spin when landing after a glide
				if speed > 5*mo.scale
				and not CBW_Battle
					player.pflags = $ &~ PF_SPINDOWN
				end
				
				-- trigger landing lag in battle
				mo.raylandinglag = (CBW_Battle) and 1 or 0
			elseif not player.glidetime and not mo.rayclingglidetime
				player.raygroundcut = false
			end
		end
	end
	
	-- remove this if needed
	if (mo.raylandinglag == 2)
	and not (P_IsObjectOnGround(mo))
		if not ((player.pflags & PF_JUMPED) and (player.pflags & PF_THOKKED))
			mo.raylandinglag = nil
		end
	end
	
	-- trigger landing lag
	if (mo.eflags & MFE_JUSTHITFLOOR)
		local source_landing = false
		if (player.raywasthokked and not player.raywasjush and (mo.floorz != mo.watertop) and mo.state == S_PLAY_FALL)
			source_landing = true
		end
		
		if (mo.raylandinglag)
		and not (source_landing)
			Ray.TriggerLandingLag(player)
			mo.raylandinglag = nil
		end
	end
	player.raywasthokked = (player.pflags & PF_THOKKED)
	player.raywasjush = (player.pflags & (PF_JUMPED|PF_SHIELDABILITY))
	if (P_IsObjectOnGround(mo)) then mo.raylandinglag = nil end	-- make sure this is removed
end