-- Ray's wall cling
-- Adapted from Mighty's wall kick by Snu

-- constants
Ray.WallClingJumpSpeed = FRACUNIT*10
Ray.WallClingJumpHeight = FRACUNIT*8
Ray.WallClingJumpMaxAngle = 64
Ray.WallClingMomStabilizer = FRACUNIT*6/9
Ray.WallClingMomZFallingStabilizer = FRACUNIT/3
Ray.WallClingSlideTics = 4*TICRATE	-- how long you can stay clinged before you start sliding down
Ray.WallClingSlideModifier = FRACUNIT*94/100
Ray.WallClingMinXYSlide = FRACUNIT*4/9
Ray.WallClingSlideClingSpeed = FRACUNIT/2	-- while going up or clinging
Ray.WallClingSlideFallSpeed = FRACUNIT/6	-- when decending
Ray.WallClingDelay = TICRATE/3
Ray.WallClingDustTicBounds = {3, 12}
Ray.WallClingDustMomBounds = {FRACUNIT*3/4, 11*FRACUNIT}
Ray.WallClingSuperModifier = FRACUNIT*6/5	-- increased strength for super as a bonus
Ray.WallClingMinInfoSave = 2	-- the minimum amount of info needed to save a cling
Ray.WallClingMaxInfoSave = 3	-- the max amount of info that can be saved
Ray.WallClingMaxFrameTics = TICRATE/3
Ray.WallClingMaxFrameSpeed = FRACUNIT*12

-- functions
Ray.ResetWallCling = function(player, resetangle)
	local mo = player.mo
	
	player.raycling, mo.raycling = nil, nil
	mo.rayclingmomentum = nil
	mo.rayclingdusttics = nil
	mo.rayclingsavedmomentum = nil
	mo.rayclingoldinfo = nil
	mo.rayclingframe = nil
	mo.rayclingframetics = nil
	mo.rayclingscroller = nil
	mo.rayprevclimbing = nil
	mo.rayclingglidetime = nil
end

-- angtoint
Ray.AngToInt = function(a)
	if not a return 0 end
	a = $/ANG1
	if a < 0
		a = $+360
	end
	return a
end

-- I am SICK of doing this so many times, laziness time
Ray.WaterMom = function(mo, momz, modifier)
	modifier = $ or FRACUNIT/2
	return FixedMul(momz, (mo.eflags & MFE_UNDERWATER) and modifier or FRACUNIT)
end

-- function to check lines
Ray.CheckLine = function(player, line, allownoclimb)
	local mo = player.mo
	
	-- really easy one, dont cling to noclimb walls
	if not (allownoclimb)
		if (line.flags & ML_NOCLIMB) then return false end
	end
	
	-- horizon line
	if (line.special == 41) then return end
	
	-- sector checks
	if not (line.backsector) then return true end	-- apparently in these situations source always returns true?
	local sec = (line.backsector and not P_PointOnLineSide(mo.x, mo.y, line)) and line.backsector or line.frontsector
	
	local canclimb = false
	local checkz = {mo.z, mo.z + mo.height}
	local floorz = P_GetZAt(sec.f_slope, mo.x, mo.y, sec.floorheight)
	local ceilingz = P_GetZAt(sec.c_slope, mo.x, mo.y, sec.ceilingheight)
	
	-- floor check
	if (checkz[2] < floorz)
		canclimb = true
	end
	
	-- ceiling checks
	local thokbarrier = (floorz == ceilingz) -- floor is the same as ceiling, thok barrier?
	if P_CheckSkyHit then thokbarrier = P_CheckSkyHit(mo, line) end -- use vanilla thok barrier check instead
	if ((checkz[1] > ceilingz) and not thokbarrier) 
		canclimb = true
	end
	
	-- FOF checks
	for rover in sec.ffloors()
		if rover.flags & FF_EXISTS
		and rover.flags & FF_BLOCKPLAYER
		and checkz[1] > P_GetZAt(sec.b_slope, mo.x, mo.y, rover.bottomheight)
		and checkz[2] < P_GetZAt(sec.t_slope, mo.x, mo.y, rover.topheight)
			canclimb = true
			break
		end
	end
	
	-- if we're here, line checks went right
	return canclimb
end

-- cling onto walls after a glide
Ray.MobjMoveBlocked = function(mo, thing)
	local player = mo.player
	if not Ray.IsRay(player) or Ray.CantMove(player)
	or (thing and thing.valid
	and thing.z - thing.scale <= mo.z + mo.height
	and thing.z + thing.height + thing.scale >= mo.z)
	or mo.raycling or mo.rayclingdelay or player.climbing
	or not player.glidetime
	or not (player.cmd.buttons & BT_JUMP)
	or not (player.pflags & PF_JUMPED)
	or not (player.pflags & PF_THOKKED)
	or player.actionstate -- battlemod support
	or (CBW_Battle and player.exhaustmeter == 0) -- we're exhausted (battle mod)
	or player.gotflag or player.gotcrystal -- we have the flag (or the crystal in battlemod)
	then return end
	
	-- get momentum
	local momx, momy = mo.rayoldmomx or mo.momx, mo.rayoldmomy or mo.momy
	local momz = mo.momz
	
	-- try to get a line to climb
	local line = player.lastlinehit
	local side = player.lastsidehit
	P_SlideMove(mo)
	if player.lastlinehit == -1
		player.lastlinehit = line
		player.lastsidehit = side
		line = nil
	else
		line = lines[player.lastlinehit]
	end
	
	if line and line.valid
	and not (player.cmd.buttons & BT_SPIN)
	and Ray.CheckLine(player, line) -- make sure everything is good here
		-- store momentum
		local wallangle = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y)
		local moangle = R_PointToAngle2(0, 0, momx, momy)
		local hitangle = wallangle - moangle
		local speedfactor = abs(cos(hitangle))
		local speed = FixedMul(speedfactor, FixedHypot(momx, momy))
		
		local reverseangle = (wallangle - ANGLE_90) - moangle
		if (Ray.AngToInt(reverseangle) > 180)
			speed = $ * -1
		end
		speed = FixedMul(Ray.WallClingMomStabilizer, speed)
		
		if ((momz * P_MobjFlip(mo)) < 0)
			momz = FixedMul(Ray.WallClingMomZFallingStabilizer, momz)
		end
		mo.rayclingmomentum = {speed, momz}
		mo.rayclingglidetime = player.glidetime
		
		-- face the wall
		local whichside = P_PointOnLineSide(mo.x, mo.y, line)
		local angle = wallangle + (ANGLE_90 * (whichside and -1 or 1))
		player.drawangle = angle
		
		-- little hack to face the wall without moving the camera (SMS Alfredo)
		local thrustx = P_ReturnThrustX(mo, angle, player.height)
		local thrusty = P_ReturnThrustY(mo, angle, player.height)
		local thok = P_SpawnMobjFromMobj(mo, thrustx, thrusty, 0, MT_THOK)
			
		local oldtarget = mo.target
		mo.target = thok
		A_FaceTarget(mo)
			
		mo.target = oldtarget
		thok.state = S_NULL
		
		-- cling to the wall!
		player.climbing = 5
		player.raycling, mo.raycling = true, true	-- trust me there's a reason we have two saved
		mo.state = (speed or momz) and S_PLAY_CLIMB or S_PLAY_CLING
		mo.rayclingdusttics = Ray.WallClingDustTicBounds[2]
		S_StopSoundByID(mo, sfx_s3k4a)
		S_StartSound(mo, sfx_kc45)
		
		-- setup frames
		if (mo.state == S_PLAY_CLIMB)
			mo.rayclingframe = 0
			mo.rayclingframetics = Ray.WallClingMaxFrameTics
		end
			
		-- reset stuff on cling
		player.pflags = $ & ~(PF_JUMPED|PF_NOJUMPDAMAGE|PF_STARTJUMP|PF_SPINNING|PF_THOKKED|PF_SHIELDABILITY)
		player.secondjump = 0
		//if (player.glidetime) then player.glidetime = min(1, $) end	-- spin puff reset?
		
		-- reduce glide degrade
		mo.rayclingcount = $ and $+1 or 1
		if (player.rayglidepower == nil) then player.rayglidepower = Ray.GLIDEMAXPOWER end
		//print("glide power: "..player.rayglidepower)
		
		local total_degrade = max(0, Ray.GLIDEMAXPOWER - player.rayglidepower)	-- just using the max to be safe, who knows
		local degrade_remove = total_degrade / mo.rayclingcount
		player.rayglidepower = min(Ray.GLIDEMAXPOWER, $ + degrade_remove)	-- even though it's called _remove, add it back
		//print("count: "..mo.rayclingcount)
		//print("restored glide power: "..player.rayglidepower)
		
		-- dust effect
		local speed = FRACUNIT*7/2
		local amount = 8
		local diff = 360/amount
		
		local angle = mo.angle
		local horzangle = angle + ANGLE_90
		for i = 1, amount
			local vertangle = (i*diff)*ANG1
			
			-- spawn the dust
			local distance = FRACUNIT*13
			local xoff = FixedMul(cos(angle), distance)
			local yoff = FixedMul(sin(angle), distance)
			local zoff = FixedMul(FRACUNIT/2, mo.height)
					
			local mobj = P_SpawnMobjFromMobj(mo, xoff, yoff, zoff, MT_THOK)
			mobj.target = mo
			
			-- horizontal
			local horizontal_thrust = FixedMul(sin(vertangle), speed)
			P_InstaThrust(mobj, horzangle, FixedMul(mo.scale, horizontal_thrust))
			mobj.angle = horzangle
			
			-- vertical
			local vertical_thrust = FixedMul(cos(vertangle), speed)
			P_SetObjectMomZ(mobj, vertical_thrust)
			
			-- set state under the right conditions
			mobj.scale = FixedMul(FRACUNIT*2/3, $)
			mobj.state = S_SPINDUST2
			
			if (mo.eflags & (MFE_TOUCHWATER|MFE_UNDERWATER)) 
				mobj.state = S_SPINDUST_BUBBLE1
			elseif (player.powers[pw_shield] == SH_ELEMENTAL)
				mobj.state = S_SPINDUST_FIRE1
			end
		end
	
	-- bump off
	elseif mo.state == S_PLAY_GLIDE
		mo.state = S_PLAY_JUMP
		player.pflags = $ &~ PF_NOJUMPDAMAGE
	end
	
	player.powers[pw_pushing] = 3
	return true
end

-- prevent movement while clinged as Ray!
Ray.PreThinkFrame = function()
	for player in players.iterate
		local mo = player.mo
		if not ((mo and mo.valid) and mo.health) then continue end
		if not (mo.raycling) then continue end
		
		-- prevent movement, but store forwardmove/sidemove values
		local cmd = player.cmd
		player.rayforwardmove = cmd.forwardmove
		player.raysidemove = cmd.sidemove
		
		cmd.forwardmove = 0
		cmd.sidemove = 0
	end
end

-- jump off the wall we're clinged to
Ray.JumpSpecial = function(player)
	local mo = player.mo
	if not Ray.IsRay(player)
	or Ray.CantMove(player)
	or not player.climbing
	or not mo.raycling
	or player.pflags & PF_JUMPDOWN
	then return end
	
	-- get jump angle (make sure it's restricted too)
	local maxangle = Ray.WallClingJumpMaxAngle	-- max walljump angle
	local cameraangle = (player.cmd.angleturn * FRACUNIT)
	local camerachange = true
	local anglediff = mo.angle - cameraangle
	
	-- account for wall kick modes
	if (player.pflags & PF_ANALOGMODE)	-- not fully camera based
		if player.rayforwardmove or player.raysidemove
			local inputangle = R_PointToAngle2(0, 0, player.rayforwardmove*FRACUNIT, -player.raysidemove*FRACUNIT)
			anglediff = $ - inputangle
			if abs(anglediff) > ANGLE_90 or anglediff == ANGLE_180
				anglediff = InvAngle($) + ANGLE_180
			end
			
			camerachange = false -- no camera change!
		else
			-- inputting nothing jumps off the wall directly with a camera change
			anglediff = 0
		end
	end
	
	anglediff = max(-(maxangle), min(maxangle, $ / ANG1)) * ANG1	-- min/max it
	local angle = mo.angle + ANGLE_180 + anglediff
	local anglediff = angle - cameraangle
	
	local jumpspeed = FixedMul(player.jumpfactor, FixedMul(mo.scale, Ray.WallClingJumpSpeed))
	local jumpheight = FixedMul(player.jumpfactor, Ray.WallClingJumpHeight)
	if (P_PlayerFullbright(player))
		-- increase power for super forms!
		jumpspeed = FixedMul(Ray.WallClingSuperModifier, $)
		jumpheight = FixedMul(Ray.WallClingSuperModifier, $)
	end
	
	P_DoJump(player, true)
	
	if camerachange
	or abs(anglediff) > ANGLE_90 or anglediff == ANGLE_180
		mo.angle = angle
	else
		mo.angle = cameraangle
	end
	player.drawangle = angle
	
	P_InstaThrust(mo, angle, Ray.WaterMom(mo, jumpspeed))
	P_SetObjectMomZ(mo, Ray.WaterMom(mo, jumpheight))
	S_StartSound(mo, sfx_jump)
	
	-- reset wallcling vars
	Ray.ResetWallCling(player)
	return true
end

-- ThinkFrame for cling
Ray.WallCling = function(player)
	local mo = player.mo
	if not (mo and mo.valid) then return end
	if (not Ray.IsRay(player) or Ray.CantMove(player))
	then Ray.ResetWallCling(player) return end
	
	-- reset stuff on the floor
	if (P_IsObjectOnGround(mo) or mo.rayclingtics == nil)
		mo.rayclingdelay = Ray.WallClingDelay
		mo.rayclingtics = Ray.WallClingSlideTics
		mo.rayclingoldz = nil
		mo.rayclingcount = 0
	else
		mo.rayclingdelay = $ and $-1
	end
	
	if (player.climbing)
		-- if we're not super, lose our grip!
		if not (P_PlayerFullbright(player))
			mo.rayclingtics = $ - 1
		else
			mo.rayclingtics = Ray.WallClingSlideTics
		end
		
		-- keep glidetime when clinging
		if mo.rayclingglidetime != nil
			player.glidetime = mo.rayclingglidetime
			if player.climbing == 1
				mo.rayclingglidetime = nil
			end
		end
		
		-- store old position (trust me this'll be useful)
		if (mo.rayclingtics > 0)	-- gotta still be able to cling
		and (mo.rayclingmomentum[1] or mo.rayclingmomentum[2])	-- gotta have some momentum
			mo.rayclingoldinfo = $ or {}
			
			local info = {
				pos = {mo.x, mo.y, mo.z},
				lastlinehit = player.lastlinehit,
			}
			table.insert(mo.rayclingoldinfo, info)
			
			-- prevent the table from getting too big
			if (#mo.rayclingoldinfo > Ray.WallClingMaxInfoSave)
				table.remove(mo.rayclingoldinfo, 1)
			end
		elseif mo.rayclingscroller and mo.rayclingoldinfo
			mo.rayclingoldinfo = nil -- get rid of it if we don't need it
		end
	
	elseif (player.raycling)
		-- this is why we have it saved in mo as well
		if (mo.raycling)
			if not (P_IsObjectOnGround(mo))	
				if (player.cmd.buttons & BT_SPIN)
					-- minimum height
					if ((mo.eflags & MFE_VERTICALFLIP) and mo.momz > FixedMul(mo.scale, -3*FRACUNIT))
					or (not (mo.eflags & MFE_VERTICALFLIP) and mo.momz < FixedMul(mo.scale, 3*FRACUNIT))
						mo.momz = FixedMul(mo.scale, FRACUNIT*3) * P_MobjFlip(mo)
					end
				else
					-- set momentum
					if (mo.rayclingsavedmomentum)
						local moms = {"momx", "momy", "momz"}
						for i = 1, 3
							if not (mo.rayclingsavedmomentum[i]) then continue end
							local mom = moms[i]
							mo[mom] = mo.rayclingsavedmomentum[i]
						end
					end
					
					mo.rayclingdelay = Ray.WallClingDelay	-- this helps more than you'd think
				end
			end
			
			S_StartSound(mo, sfx_kc45)
		end
		Ray.ResetWallCling(player)
	end
end

-- playerthink lets us edit momentum while climbing
Ray.ClingThinker = function(player)
	local mo = player.mo
	
	-- wall cling checks
	if (not Ray.IsRay(player) or Ray.CantMove(player)) then return end
	if not (mo.raycling)
		mo.rayoldmomx, mo.rayoldmomy = mo.momx, mo.momy	-- i hate that I have to store this...
	end
	
	if not (player.climbing)
		-- gotta do this here, since we can INSTANTLY restick and prevent camera shit
		if not (player.cmd.buttons & BT_SPIN)
		and (mo.rayclingoldinfo)
		and (#mo.rayclingoldinfo >= Ray.WallClingMinInfoSave)	-- got enough to safely do this
		and (mo.rayclingtics > 0)	-- we're sliding off!
		and not (mo.rayclingscroller) -- not on a scroller!
			-- go back to old position
			local info = mo.rayclingoldinfo[1]	-- this'll be the oldest position
			P_SetOrigin(mo, info.pos[1], info.pos[2], info.pos[3])
							
			-- remove momentum
			mo.momx, mo.momy, mo.momz = 0, 0, 0
			mo.rayclingmomentum[1] = 0
			mo.rayclingmomentum[2] = 0
							
			-- face the wall
			local line = lines[info.lastlinehit]
							
			local whichside = P_PointOnLineSide(mo.x, mo.y, line)
			local angle = R_PointToAngle2(line.v1.x, line.v1.y, line.v2.x, line.v2.y)
			angle = $ + (ANGLE_90 * (whichside and -1 or 1))
			player.drawangle = angle
							
			-- little hack to face the wall without moving the camera (SMS Alfredo)
			local thrustx = P_ReturnThrustX(mo, angle, player.height)
			local thrusty = P_ReturnThrustY(mo, angle, player.height)
			local thok = P_SpawnMobjFromMobj(mo, thrustx, thrusty, 0, MT_THOK)
								
			local oldtarget = mo.target
			mo.target = thok
			A_FaceTarget(mo)
								
			mo.target = oldtarget
			thok.state = S_NULL
											
			-- stay clinged
			player.climbing = 5
			player.raycling, mo.raycling = true, true	-- trust me there's a reason we have two saved
			mo.state = S_PLAY_CLING
			mo.rayclingdusttics = Ray.WallClingDustTicBounds[2]
			mo.rayclingglidetime = player.glidetime
							
			-- delete old info, not needed anymore
			mo.rayclingoldinfo = nil
		else
			-- nvm, cease functions
			return
		end
	end
	
	-- remove horizontal momentum then set it
	local removemomxy = abs(mo.rayclingmomentum[1]) - FixedMul(Ray.WallClingSlideModifier, abs(mo.rayclingmomentum[1]))
	if (removemomxy < FixedMul(mo.scale, Ray.WallClingMinXYSlide))
		-- would just use max() but it'd be too long on one line and this is just about the same thing
		removemomxy = FixedMul(mo.scale, Ray.WallClingMinXYSlide)
	end
		
	local minmax_func = max
	if (mo.rayclingmomentum[1] < 0)
		removemomxy = $ * -1
		minmax_func = min
	end
	mo.rayclingmomentum[1] = minmax_func(0, $ - (removemomxy / ((mo.eflags & MFE_UNDERWATER) and 2 or 1)))
		
	local line = lines[player.lastlinehit]
	if (line and line.valid)	-- we need this
		-- if we have momentum past the first few frames of climbing, we're probably scrolling
		mo.rayclingscroller = mo.rayprevclimbing == 1 and player.climbing == 1
		and (mo.momx or mo.momy or mo.momz) or 0
		
		local wallangle = R_PointToAngle2(line.v2.x, line.v2.y, line.v1.x, line.v1.y)
		P_Thrust(mo, wallangle, mo.rayclingmomentum[1])	-- this currently applies for scrollers, too good!
	elseif (mo.rayclingscroller)
		mo.rayclingscroller = 0
	end
	mo.rayprevclimbing = player.climbing
	
	-- first calculate slide speed
	local slidespeed = Ray.WallClingSlideClingSpeed
	if (mo.eflags & MFE_GOOWATER)
		if ((mo.rayclingmomentum[2] * P_MobjFlip(mo)) > 0)
			slidespeed = Ray.WallClingSlideFallSpeed
		end
	else
		if ((mo.rayclingmomentum[2] * P_MobjFlip(mo)) < 0)
			slidespeed = Ray.WallClingSlideFallSpeed
		end
	end
	slidespeed = FixedMul(mo.scale, $)
		
	-- now remove it from our momz
	if (mo.rayclingtics > 0 or mo.rayclingscroller)	-- we can still cling to the wall without penalty!
		if (mo.rayclingmomentum[2])	-- we still have momentum! remove it quickly
			-- calculate how much to reduce momz by
			local removemomz = FixedMul(mo.scale, Ray.WallClingSlideClingSpeed)
			local minmax_func = max
			-- flip these if we need to increase to 0
			if (mo.rayclingmomentum[2] < 0)
				removemomz = $ * -1
				minmax_func = min
			end
			
			mo.rayclingmomentum[2] = minmax_func(0, $ - (removemomz / ((mo.eflags & MFE_UNDERWATER) and 2 or 1)))
		end
	else
		local removemomz = slidespeed * P_MobjFlip(mo)
		if (mo.eflags & MFE_GOOWATER)
			removemomz = $ * -1
		end
		mo.rayclingmomentum[2] = $ - (removemomz / ((mo.eflags & MFE_UNDERWATER) and 2 or 1))
	end
	
	-- now set it
	mo.momz = $ + mo.rayclingmomentum[2]
	
	-- save these for launching off the wall...
	mo.rayclingsavedmomentum = {mo.momx, mo.momy, mo.momz}
	
	-- visuals!
	local speed = FixedHypot(mo.rayclingmomentum[2], mo.rayclingmomentum[1])	-- take both momxy and momz into account
	
	if (speed)
		-- frame animation
		local frame_speed = FixedMul(mo.scale, Ray.WallClingMaxFrameSpeed)
		local frame_tics = Ray.WallClingMaxFrameTics
		local maxtics = max(1, frame_tics - ((speed * frame_tics) / frame_speed))
		mo.rayclingframetics = $ and min(maxtics, $) or maxtics
		
		local numframes = 2
		mo.rayclingframetics = $-1
		if not (mo.rayclingframetics)
			mo.rayclingframe = $+1
				
			-- reset frame after max frame is reached
			if (mo.rayclingframe > (numframes - 1))
				mo.rayclingframe = 0
			end
		end
		
		mo.state = S_PLAY_CLIMB
		mo.frame = ($ &~ FF_FRAMEMASK) | mo.rayclingframe
	
		-- dust particles on the wall
		local mombounds = Ray.WallClingDustMomBounds
		if (speed > mombounds[1])
			speed = $ - mombounds[1]
			
			-- duration is based on how fast we're going
			local momrange = FixedMul(mo.scale, mombounds[2] - mombounds[1])
			local ticbounds = Ray.WallClingDustTicBounds
			local ticrange = ticbounds[2] - ticbounds[1]
			
			local mintics = max(ticbounds[1], ticbounds[2] - ((speed * ticrange) / momrange))
			mo.rayclingdusttics = min(mintics, $) - 1
			if not (mo.rayclingdusttics)
				for i = 1, 2	-- for both shoes
					for j = 1, 2
						-- distance to the wall
						local distance = 13
						local xoff = distance*cos(mo.angle)
						local yoff = distance*sin(mo.angle)
								
						-- distance to espio's hand (with some randomization)
						local angle = mo.angle - ((i == 1) and ANGLE_270 or ANGLE_90)
						local distance = 12 + P_RandomRange(-8, 8)
						xoff = $ + distance*cos(angle)
						yoff = $ + distance*sin(angle)
						local zoff = FRACUNIT*((i == 1) and 6 or 3) + (P_RandomRange(-8, 8)*FRACUNIT)
								
						local dust = P_SpawnMobjFromMobj(mo, xoff, yoff, zoff, MT_SPINDUST)
						dust.state = (mo.eflags & MFE_UNDERWATER) and S_SPINDUST_BUBBLE1 or $
						dust.scale = $*2/3
						P_InstaThrust(dust, R_PointToAngle2(0, 0, mo.momx, mo.momy), FixedMul(FRACUNIT/4, -(FixedHypot(mo.momx, mo.momy))))
						dust.momz = FixedMul(FRACUNIT/4, mo.momz * -1)
						S_StartSound(mo, sfx_s3k7e)
					end
				end
				
				-- reset tics
				mo.rayclingdusttics = mintics
			end
		end
	else
		-- reset frame info
		mo.rayclingframe = 0
		mo.rayclingframetics = Ray.WallClingMaxFrameTics
	end
end